// $Id: SmithWaterman.cc,v 1.2 2005/10/04 00:18:39 vern Exp $
//
// Copyright (c) 2005
//      The Regents of the University of California.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that: (1) source code distributions
// retain the above copyright notice and this paragraph in its entirety, (2)
// distributions including binary code include the above copyright notice and
// this paragraph in its entirety in the documentation or other materials
// provided with the distribution, and (3) all advertising materials mentioning
// features or use of this software display the following acknowledgement:
// ``This product includes software developed by the University of California,
// Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
// the University nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

#include "config.h"

#include <algorithm>
#include <ctype.h>

#include "SmithWaterman.h"
#include "Var.h"

void BroSubstring::operator=(const BroSubstring &bst)
	{
	_string = bst._string;
	_index1 = bst._index1;
	_index2 = bst._index2;
	_size = bst._size;
	_new = bst._new;
	}

bool BroSubstring::operator<(const BroSubstring &bst) const
	{
	if (_index1 < bst._index1)
		return true;
	if (_index1 > bst._index1)
		return false;

	if (_index2 < bst._index2)
		return true;
	if (_index2 > bst._index2)
		return false;

	return false;
	}

VectorVal* BroSubstring::VecToPolicy(Vec* vec)
	{
	RecordType* sw_substring_type =
		internal_type("sw_substring")->AsRecordType();
	if ( ! sw_substring_type )
		return 0;

	VectorVal* result =
		new VectorVal(internal_type("sw_substring_vec")->AsVectorType());
	if ( ! result )
		return 0;

	for ( unsigned int i = 0; i < vec->size(); ++i )
		{
		BroSubstring& bst = (*vec)[i];

		RecordVal* st_val = new RecordVal(sw_substring_type);
		st_val->Assign(0, new StringVal(bst.GetString()->Len(),
				(const char *) bst.GetString()->Bytes()));
		st_val->Assign(1, new Val(bst.GetIndex1(), TYPE_COUNT));
		st_val->Assign(2, new Val(bst.GetIndex2(), TYPE_COUNT));
		st_val->Assign(3, new Val(bst.IsNewAlignment(), TYPE_BOOL));
		result->Assign(i+1, st_val, 0);
		}

	return result;
	}

BroSubstring::Vec* BroSubstring::VecFromPolicy(VectorVal* vec)
	{
	Vec* result = new Vec();

	// VectorVals start at index 1!
	for ( unsigned int i = 1; i <= vec->Size(); ++i )
		{
		Val* v = vec->Lookup(i);	// get the RecordVal
		if ( ! v )
			continue;

		const BroString* string = v->AsRecordVal()->Lookup(0)->AsString();
		int index1 = v->AsRecordVal()->Lookup(1)->AsCount();
		int index2 = v->AsRecordVal()->Lookup(2)->AsCount();
		bool new_alignment = v->AsRecordVal()->Lookup(3)->AsBool();

		BroSubstring substr(*string, index1, index2);
		substr.MarkNewAlignment(new_alignment);

		result->push_back(substr);
		}

	return result;
	}

char* BroSubstring::VecToString(Vec* vec)
	{
	string result("[");

	for ( BroSubstring::VecIt it = vec->begin(); it != vec->end(); ++it )
		{
		result += it->GetString()->ExpandedString();
		result += ",";
		}

	result += "]";
	return strdup(result.c_str());
	}


// A node in Smith-Waterman's dynamic programming matrix.  Each node
// contains the byte it represents in the case of a match, the score
// at this point, and a pointer to the previous cell. Previous means
// one up and left in case of a match, or a jump somewhere above and
// left in case of a gap.
//
struct SWNode {
	// ID field for the cell, for debugging purposes.
	int id;

	u_char swn_byte;
	bool swn_byte_assigned;

	// The score in this cell. The cell with the globally best score
	// marks the end of the alignment.
	int swn_score;

	// Pointer to previous match, walking back yields subsequence.
	SWNode* swn_prev;
};


// Quick helper function that calculates the coordinates of a node in the
// matrix via pointer arithmetics.
static void sw_get_node_indices(SWNode* base, SWNode* node, int len,
				int& row, int& col)
	{
	int offset = node - base;
	col = offset % len;
	row = offset / len;
	}

// Helper function returning the highest-scoring node in a given column.
// If no node in the column has a score > 0, null is returned.
static SWNode* sw_get_col_max_node(SWNode* matrix, int height,
					int width, int col)
	{
	// Start with top node and go down.
	SWNode* node = &matrix[col];

	for ( int i = 1; i < height; ++i )
		{
		SWNode* col_node = &matrix[i * width + col];
		if ( col_node->swn_score > node->swn_score )
			node = col_node;
		}

	return node->swn_score > 0 ? node : 0;
	}

BroSubstring::Vec* smith_waterman(const BroString* s1, const BroString* s2,
					SWParams& params)
	{
	BroSubstring::Vec* result = new BroSubstring::Vec();

	if ( ! s1 || ! s2 )
		return result;

	byte_vec string1 = s1->Bytes();
	byte_vec string2 = s2->Bytes();

	// Length of both strings, plus one because SW needs an extra row
	// and column.
	int len1 = s1->Len() + 1;
	int len2 = s2->Len() + 1;

	int counter = 1;

	SWNode matrix[len1][len2];	// dynamic programming matrix
	SWNode* node_max = 0;	// pointer to the best score's node
	SWNode* node_br_max = 0;	// pointer to lowest-right matching node

	// The highest score in the matrix, globally. We initialize to 1
	// because we are only interested in real scores (initializing to
	// -infty would mean 0 is larger and would complicate the link
	// structure in the matrix).
	int matrix_max = 1;
	int br_max_r = 0;
	int br_max_b = 0;


	// Matrix initialization.

	memset(&matrix, 0, sizeof(SWNode) * len1 * len2);

	// Assign IDs to each cell -- this is only for debugging purposes
	// and can go later.
	int i, j;
	for ( i = 1; i < len1; ++i )
		for ( j = 1; j < len2; ++j )
			matrix[i][j].id = counter++;


	// Subsequence calculation.

	int row, col;

	for ( i = 1; i < len1; ++i )
		{
		for ( j = 1; j < len2; ++j )
			{
			// Current node, top/left neighbours.
			//
			SWNode* current = &matrix[i][j];
			SWNode* node_tl = &matrix[i-1][j-1];
			SWNode* node_l = &matrix[i][j-1];
			SWNode* node_t = &matrix[i-1][j];

			// Scores of neighboring nodes.
			//
			int score_t = node_t->swn_score;
			int score_l = node_l->swn_score;
			int score_tl = node_tl->swn_score;

			// If strings at current indices match, assign new
			// score to current node. Minus-one adjustments
			// are necessary since the matrix has one extra row
			// and column.
			//
			if ( string1[i-1] == string2[j-1] )
				{
				// We have a match: improve previous score.
				score_tl += 1;

				// If we're continuing a chain of matches,
				// rate higher. This favours longer
				// consecutive substrings.
				//
				if ( node_tl->swn_byte_assigned )
					score_tl += 99;

				// Store the byte we've matched in the node for
				// easier access.
				//
				current->swn_byte = string1[i-1];
				}

			// Pick the score among the neighbors that is now
			// highest.  This is the core of Smith-Waterman.
			//
			current->swn_score =
				max(max(score_t, score_l), score_tl);

			// Establish predecessor chain according to neighbor
			// with best score.
			//
			if ( current->swn_score == score_tl )
				{
				// If we had matched bytes (*and* it's the
				// best neighbor), mark the node accordingly.
				//
				if ( string1[i-1] == string2[j-1] )
					{
					current->swn_byte_assigned = true;

					if ( i >= br_max_b && j >= br_max_r )
						{
						node_br_max = current;
						br_max_b = i;
						br_max_r = j;
						}
					}

				current->swn_prev = node_tl;
				}

			else if ( current->swn_score == score_l )
				current->swn_prev = node_l;

			else
				current->swn_prev = node_t;

			// Check if we have a new global maximum -- we
			// specifically track the node that is the global
			// maximum so we now from where to backtrack at
			// the end of the matrix iteration.
			//
			if ( current->swn_score >= matrix_max )
				{
				node_max = current;
				matrix_max = current->swn_score;
				}

#if 0
			printf("%4i/%.5i%c/%.5i[%c%c] ",
				current->swn_score != node_t->swn_score ? current->swn_score : 0,
				current->id,
				current->swn_byte_assigned ? '*' : ' ',
				current->swn_prev ? current->swn_prev->id : 0,
			       string1[i-1], string2[j-1]);
			printf("%.5i ", current->swn_score);
#endif
			}

#if 0
		printf("\n");
#endif
		}


	// Result generation.

	// How we do this depends on the mode we operate in.  In SW_SINGLE, we
	// follow the path from the best node until there is no predecessor
	// (that is, when we hit a node in row 0), and stop.  In SW_MULTIPLE,
	// ### remainder of comment text was truncated - VP

	// Result string, with a pointer to fill it since the string is
	// traversed in reverse order.
	string substring = "";
	SWNode* node = 0;	// node for backward iteration

	if ( params._sw_variant == SW_MULTIPLE )
		node = node_br_max;
	else
		node = node_max;

	// printf("WALKING from %i/%i\n", node->id, matrix_max);

	while ( node )
		{
		// printf("NODE: %i\n", node->id);

		// Once we hit a gap, terminate the string and prepend
		// it to our result vector, IF it has at least the length
		// requested through the params._min_toklen parameter.
		//
		if ( node->swn_byte_assigned )
			{
			sw_get_node_indices(&matrix[0][0], node, len2, row, col);
			substring += node->swn_byte;
			// printf("SUBSTRING: %s\n", substring.c_str());
			}
		else
			{
			// printf("GAP\n");
			if ( substring.size() >= params._min_toklen )
				{
				reverse(substring.begin(), substring.end());
				BroSubstring bst(substring, row - 1, col - 1);
				result->push_back(bst);
				}

			substring = "";
			}

		if ( params._sw_variant == SW_MULTIPLE && ! node->swn_prev )
			{
			int tmp_row, tmp_col;
			sw_get_node_indices(&matrix[0][0], node, len2,
						tmp_row, tmp_col);

			if ( result->size() > 0 )
				result->back().MarkNewAlignment(true);

			if ( tmp_col == 0 )
				break;

			node = sw_get_col_max_node(&matrix[0][0], len1, len2,
							tmp_col - 1);
			}
		else
			node = node->swn_prev;
		}

	if ( substring.size() > 0 )
		{
		reverse(substring.begin(), substring.end());
		BroSubstring bst(substring, row - 1, col - 1);
		bst.MarkNewAlignment(true);
		result->push_back(bst);
		}

	// Vectors do not support push_front, so we need to reverse the vector
	// explicitly.
	//
	reverse(result->begin(), result->end());

	return result;
	}
