// $Id: SmithWaterman.h,v 1.2 2005/10/04 00:18:39 vern Exp $
//
// Copyright (c) 2005
//      The Regents of the University of California.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that: (1) source code distributions
// retain the above copyright notice and this paragraph in its entirety, (2)
// distributions including binary code include the above copyright notice and
// this paragraph in its entirety in the documentation or other materials
// provided with the distribution, and (3) all advertising materials mentioning
// features or use of this software display the following acknowledgement:
// ``This product includes software developed by the University of California,
// Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
// the University nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

#ifndef smith_waterman_h
#define smith_waterman_h

#include "BroString.h"
using namespace std;

// BroSubstrings are essentially BroStrings, augmented with indexing
// information required for the Smith-Waterman algorith.
//
// XXX: Could be derived from BroString, instead of aggregating one?
//
class BroSubstring {

public:
	typedef vector<BroSubstring> Vec;
	typedef Vec::iterator VecIt;
	typedef Vec::const_iterator VecCIt;

	BroSubstring(const string &string, int index1, int index2)
	: _string(string), _index1(index1), _index2(index2), _new(false)
		{ _size = string.size(); }

	BroSubstring(const BroString &string, int index1, int index2)
	: _string(string), _index1(index1), _index2(index2), _new(false)
		{ _size = string.Len(); }

	BroSubstring(const BroSubstring &bst)
	: _string(bst._string), _index1(bst._index1), _index2(bst._index2),
	  _size(bst._size), _new(bst._new)
		{ }

	void operator=(const BroSubstring &bst);
	bool operator<(const BroSubstring &bst) const;

	BroString* GetString() { return &_string; }
	int GetIndex1() const  { return _index1; }
	int GetIndex2() const  { return _index2; }
	int GetSize() const    { return _size; }

	void MarkNewAlignment(bool mark) { _new = mark; }
	bool IsNewAlignment()   { return _new; }

	// Helper methods for vectors:
	//
	static VectorVal* VecToPolicy(Vec* vec);
	static Vec* VecFromPolicy(VectorVal* vec);
	static char* VecToString(Vec* vec);

private:
	BroSubstring();

	// The (sub)string itself.
	BroString _string;

	// Start index of the string, in the 1st/2nd string passed to SW.
	int _index1;
	int _index2;

	// The size of the string.
	int _size;

	 // True if this node marks the start of a new alignment.
	bool _new;
};



// Smith-Waterman Implementation
// ---------------------------------------------------------------------
//

// We support two modes of operation: finding a single optimal alignment,
// and repeated alignments.
//
enum SWVariant {
	SW_SINGLE   = 0,	// return a single, optimum alignment
	SW_MULTIPLE = 1,	// find repeated, non-overlapping alignments
};

// Parameters for Smith-Waterman are stored in this simple record.
//
struct SWParams {
	SWParams(unsigned int min_toklen = 3, SWVariant sw_variant = SW_SINGLE)
		{
		_min_toklen = min_toklen;
		_sw_variant = sw_variant;
		}

	// The minimum string size to report.  For example, min_toklen = 2
	// won't report any common single-letter subsequences.
	unsigned int _min_toklen;

	SWVariant _sw_variant;
};


// The smith_waterman() algorithm finds the longest common subsequence(s)
// of two strings, also known as the best local alignment.  A subsequence
// is a sequence of common substrings.
//
//  s1:         first input string
//  s2:         second input string
//  params:     Smith-Waterman parameters.
//
// Subsequences of a string are any strings based on the original one
// with individual characters left out. Note that this is different
// from the longest common substring problem.
//
// The function returns a vector consisting of all substrings comprising
// the subsequence.  With each string you also get the indices of both
// input strings where the string occurs.  On error, or if no common
// subsequence exists, an empty vector is returned.
//
extern BroSubstring::Vec* smith_waterman(const BroString* s1,
					 const BroString* s2,
					 SWParams& params);

#endif
